import os
import glob
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

algorithmMap = {
    "irmv1": "IRMV1",
    "infer_irmv1": "ZIN",
    "irmv1_tvl1": "IRM-TV-L1",
    "infer_irmv1_tvl1": "ZIN-TV-L1",
    "irmv1_multi_class": "IRMV1",
    "infer_irmv1_multi_class": "ZIN",
    "irmv1_multi_class_tvl1": "IRM-TV-L1",
    "infer_irmv1_multi_class_tvl1": "ZIN-TV-L1",
}
folder_path = "figure_data"
saveFormat="png"
csv_files = glob.glob(os.path.join(folder_path, "*.csv"))
for file in csv_files:
    data = pd.read_csv(file)

    file = file.replace("figure_data\\", "")
    file = file.replace(".csv", "")
    names = file.split("_")
    try:
        redRecord = data["redRecord"]
        blueRecord = data["blueRecord"]

        iter = data["redRecord"][
            data["redRecord"] != 0
        ].first_valid_index()
        redRecord = redRecord[iter:]
        blueRecord = blueRecord[iter:]
        xl = range(iter, iter + redRecord.size, 1)
        plt.figure(figsize=(15, 6), dpi=300)
        plt.plot(xl, redRecord, color="red", label="penalty")
        plt.plot(xl, blueRecord, color="blue", label="blue")
        title = f"IMAGE1 {names[0]} with {'_'.join(names[1:])} algorithm"
        plt.title(title)
        plt.xlabel("epoch")
        plt.ylabel("values")
        plt.legend()
        plt.savefig(os.path.join(folder_path, title),format=saveFormat)
        # plt.savefig(os.path.join(folder_path, title))
        print(f"save {title}")
    except Exception:
        pass
    try:
        phi_residual = data["phi_residual"][iter:]
        psi_residual = data["psi_residual"][iter:]
        # xl = range(1, phi_residual.size + 1, 1)
        plt.figure(figsize=(15, 6), dpi=300)
        plt.plot(xl, phi_residual, color="red", label="phi")
        plt.plot(xl, psi_residual, color="blue", label="psi")
        title = f"IMAGE2 {names[0]} with {'_'.join(names[1:])} algorithm"
        plt.title(title)
        plt.xlabel("epoch")
        plt.ylabel("values")
        plt.legend()
        plt.savefig(os.path.join(folder_path, title),format=saveFormat)
        # plt.savefig(os.path.join(folder_path, title))
        print(f"save {title}")
    except Exception:
        pass

    try:
        fidelity = data["fidelity"][iter:]
        redRecord = data["redRecord"][iter:]
        plt.figure(figsize=(15, 6), dpi=300)
        plt.plot(xl, redRecord, color="red", label="penalty")
        plt.plot(xl, fidelity, color="blue", label="fidelity")
        title = f"IMAGE3 {names[0]} with {'_'.join(names[1:])} algorithm"
        plt.title(title)
        plt.xlabel("epoch")
        plt.ylabel("values")
        plt.legend()
        plt.savefig(os.path.join(folder_path, title),format=saveFormat)
        # plt.savefig(os.path.join(folder_path, title))
        print(f"save {title}")
    except Exception:
        pass
