import os
from tkinter import messagebox


def create_command(searchArgs: dict):

    argSentence = ""
    for key, value in searchArgs.items():
        arg_command = f"--{key} {str(value)} "
        argSentence += arg_command
    return argSentence


PYTHON_PATH = "D:/ProgramData/Anaconda3/envs/wyc/python.exe"  # 替换成你自己环境的python
COMMAND = f"{PYTHON_PATH} main.py "
parameterMap = {
    "ZIN": {
        "adult": {
            "l2_regularizer_weight": 0.001,
            "irm_type": "infer_irmv1_multi_class",
            "dataset": "adult",
            "n_restarts": 3,
            "seed": 1,
            "hidden_dim_infer": 64,
            "hidden_dim": 32,
            "penalty_weight_anneal": 1000,
            "lr": 0.01,
            "lr2": 0.003,
            "aux_num": 6,
            "z_class_num": 4,
            "envs_num_train": 2,
            "envs_num_test": 4,
            "penalty_anneal_iters": 950,
            "steps": 1000,
        },
        "celebaz_feature": {  # 默认参数 不想每列都写的就改默认值
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.2,
            "cons_train": "0.999_0.9",
            "cons_test": "0.01_0.2_0.8_0.999",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 39996,
            "data_num_test": 20000,
            "dataset": "celebaz_feature",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "infer_irmv1",
            "penalty_weight_anneal": 2000,
            "lr": 0.0001,
            "lr2": 0.0001,
            "penalty_anneal_iters": 2000,
            "steps": 4000,
        },
        "logit_z": {
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.1,
            "cons_train": "0.999_0.9",
            "cons_test": "0.999_0.8_0.2_0.001",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 10,
            "data_num_train": 5000,
            "data_num_test": 5000,
            "dataset": "logit_z",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "infer_irmv1",
            "penalty_weight_anneal": 50000,
            "lr": 0.005,
            "lr2": 0.0001,
            "penalty_anneal_iters": 2000,
            "steps": 4000,
        },
        "landcover": {
            "aux_num": "2",
            "batch_size": "1024",
            "seed": "112",
            "classes_num": "6",
            "dataset": "landcover",
            "opt": "adam",
            "l2_regularizer_weight": 0.001,
            "n_restarts": "1",
            "num_classes": "6",
            "z_class_num": "2",
            "noise_ratio": "0.1",
            "envs_num_train": 2,
            "cons_train": "0.999_0.7",
            "cons_test": "0.999_0.001",
            "penalty_weight": "10",
            "n_restarts": 1,
            "dim_inv": "2",
            "dim_spu": "10",
            "envs_num_test": 2,
            "irm_type": "infer_irmv1_multi_class",
            "lr": 0.1,
            "lr2": 0.001,
            "penalty_weight_anneal": 25,
            "steps": "100",
            "penalty_anneal_iters": 40,
            "scheduler": 1,
        },
    },
    "IRMV1": {
        "landcover": {
            "aux_num": "2",
            "batch_size": "1024",
            "seed": "112",
            "classes_num": "6",
            "dataset": "landcover",
            "opt": "adam",
            "l2_regularizer_weight": 0.001,
            "n_restarts": "1",
            "num_classes": "6",
            "z_class_num": "2",
            "noise_ratio": "0.1",
            "envs_num_train": 2,
            "cons_train": "0.999_0.7",
            "cons_test": "0.999_0.001",
            "penalty_weight": "10",
            "n_restarts": 1,
            "dim_inv": "2",
            "dim_spu": "10",
            "envs_num_test": 2,
            "irm_type": "irmv1_multi_class",
            "lr": "0.08",
            "lr2": "0.001",
            "penalty_weight_anneal": "1",
            "steps": "70",
            "penalty_anneal_iters": "20",
            "scheduler": 1,
        },
        "logit_z": {
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.1,
            "cons_train": "0.999_0.9",
            "cons_test": "0.999_0.8_0.2_0.001",
            "penalty_weight": 1000,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 5000,
            "data_num_test": 5000,
            "dataset": "logit_z",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "irmv1",
            "penalty_weight_anneal": 5000,
            "lr": 0.01,
            "lr2": 0.003,
            "penalty_anneal_iters": 5000,
            "steps": 8000,
        },
        "adult": {
            "l2_regularizer_weight": 0.001,
            "irm_type": "irmv1_multi_class",
            "dataset": "adult",
            "n_restarts": 3,
            "seed": 1,
            "penalty_weight_anneal": 3000,
            "lr": 0.004,
            "lr2": 0.001,
            "aux_num": 6,
            "z_class_num": 4,
            "envs_num_train": 2,
            "envs_num_test": 4,
            "penalty_anneal_iters": 1000,
            "steps": 2000,
        },
        "celebaz_feature": {  # 默认参数 不想每列都写的就改默认值
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.2,
            "cons_train": "0.999_0.8",
            "cons_test": "0.01_0.2_0.8_0.999",
            "penalty_weight": 4750,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 39996,
            "data_num_test": 20000,
            "dataset": "celebaz_feature",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "irmv1",
            "penalty_weight_anneal": 4750,
            "lr": 0.0001,
            "lr2": 0.001,
            "penalty_anneal_iters": 1000,
            "steps": 2000,
        },
    },
    "IRM-TV-L1": {
        "landcover": {
            "aux_num": "2",
            "batch_size": "1024",
            "seed": "112",
            "classes_num": "6",
            "dataset": "landcover",
            "opt": "adam",
            "l2_regularizer_weight": "0.001",
            "n_restarts": "1",
            "num_classes": "6",
            "z_class_num": "2",
            "noise_ratio": "0.1",
            "envs_num_train": 2,
            "cons_train": "0.999_0.7",
            "cons_test": "0.999_0.001",
            "penalty_weight": "10",
            "dim_inv": "2",
            "dim_spu": "10",
            "envs_num_test": 2,
            "irm_type": "irmv1_multi_class_tvl1",
            "lr": "0.08",
            "lr2": "0.001",
            "penalty_weight_anneal": "1",
            "steps": "70",
            "penalty_anneal_iters": "20",
            "scheduler": 1,
        },
        "adult": {
            "l2_regularizer_weight": 0.001,
            "irm_type": "irmv1_multi_class_tvl1",
            "dataset": "adult",
            "n_restarts": 3,
            "seed": 1,
            "penalty_weight_anneal": 1000,
            "lr": 0.01,
            "lr2": 0.004,
            "aux_num": 6,
            "z_class_num": 4,
            "envs_num_train": 2,
            "envs_num_test": 4,
            "penalty_anneal_iters": 1000,
            "steps": 1000,
        },
        "celebaz_feature": {  # 默认参数 不想每列都写的就改默认值
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.2,
            "cons_train": "0.999_0.8",
            "cons_test": "0.01_0.2_0.8_0.999",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 39996,
            "data_num_test": 20000,
            "dataset": "celebaz_feature",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "irmv1_tvl1",
            "penalty_weight_anneal": 4000,
            "lr": 0.0001,
            "lr2": 0.0005,
            "penalty_anneal_iters": 2000,
            "steps": 8000,
        },
        "logit_z": {
            "l2_regularizer_weight": 0.002,
            "noise_ratio": 0.1,
            "cons_train": "0.999_0.7",
            "cons_test": "0.999_0.8_0.2_0.001",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 5000,
            "data_num_test": 5000,
            "dataset": "logit_z",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "irmv1_tvl1",
            "penalty_weight_anneal": 4000,
            "lr": 0.02,
            "lr2": 0.0001,
            "penalty_anneal_iters": 2000,
            "steps": 4000,
        },
    },
    "ZIN-TV-L1": {
        "adult": {
            "l2_regularizer_weight": 0.001,
            "irm_type": "infer_irmv1_multi_class_tvl1",
            "dataset": "adult",
            "n_restarts": 3,
            "seed": 1,
            "penalty_weight_anneal": 1000,
            "lr": 0.01,
            "lr2": 0.003,
            "aux_num": 6,
            "z_class_num": 4,
            "envs_num_train": 2,
            "envs_num_test": 4,
            "penalty_anneal_iters": 950,
            "steps": 1000,
        },
        "celebaz_feature": {  # 默认参数 不想每列都写的就改默认值
            "l2_regularizer_weight": 0.001,
            "noise_ratio": 0.2,
            "cons_train": "0.999_0.8",
            "cons_test": "0.01_0.2_0.8_0.999",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 5,
            "data_num_train": 39996,
            "data_num_test": 20000,
            "dataset": "celebaz_feature",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "infer_irmv1_tvl1",
            "penalty_weight_anneal": 5000,
            "lr": 0.0001,
            "lr2": 0.0001,
            "penalty_anneal_iters": 2000,
            "steps": 4000,
        },
        "logit_z": {
            "l2_regularizer_weight": 0.002,
            "noise_ratio": 0.1,
            "cons_train": "0.999_0.9",
            "cons_test": "0.999_0.8_0.2_0.001",
            "penalty_weight": 2000,
            "dim_inv": 5,
            "dim_sp": 10,
            "data_num_train": 5000,
            "data_num_test": 5000,
            "dataset": "logit_z",
            "n_restarts": 1,
            "seed": 1,
            "irm_type": "infer_irmv1_tvl1",
            "penalty_weight_anneal": 22000,
            "lr": 0.005,
            "lr2": 0.0001,
            "penalty_anneal_iters": 2000,
            "steps": 100000,
        },
        "landcover": {
            "aux_num": "2",
            "batch_size": "1024",
            "seed": "112",
            "classes_num": "6",
            "dataset": "landcover",
            "opt": "adam",
            "l2_regularizer_weight": 0.001,
            "n_restarts": "1",
            "num_classes": "6",
            "z_class_num": "2",
            "noise_ratio": "0.1",
            "envs_num_train": 2,
            "cons_train": "0.999_0.7",
            "cons_test": "0.999_0.001",
            "penalty_weight": "10",
            "n_restarts": 1,
            "dim_inv": "2",
            "dim_spu": "10",
            "envs_num_test": 2,
            "irm_type": "infer_irmv1_multi_class_tvl1",
            "lr": 0.08,
            "lr2": 0.001,
            "penalty_weight_anneal": 1,
            "steps": "70",
            "penalty_anneal_iters": 10,
            "scheduler": 1,
        },
    },
}
algorithmList = ["IRMV1", "ZIN", "IRM-TV-L1", "ZIN-TV-L1"]
datasetList = ["celebaz_feature", "logit_z", "landcover", "adult"]
algorithm="ZIN"
dataset="landcover"
ARG_DEFAULT_MAP = parameterMap[algorithm][dataset]
argSentence = create_command(ARG_DEFAULT_MAP)
print(COMMAND + argSentence)
os.system(COMMAND + argSentence)

# for algorithm in algorithmList:
#     # for dataset in datasetList:
#     ARG_DEFAULT_MAP = parameterMap[algorithm][dataset]
#     argSentence = create_command(ARG_DEFAULT_MAP)
#     print(COMMAND + argSentence)
#     os.system(COMMAND + argSentence)


# searchMap = {
#     # 只写想要找的参数的值，每次会取一列的值来跑，
#     # "lr": [0.00015]*2,
#     # "lr2": [0.001]*2,
#     "steps": steps,
#     "penalty_anneal_iters": iters,
#     # "penalty_weight_anneal": [100] * 2,
# }
# n = len(list(searchMap.values())[0])
# for i in range(n):
#     print(f"current left {n-i} spaces")
#     searchArgs = ARG_DEFAULT_MAP
#     for key in searchMap.keys():
#         searchArgs[key] = searchMap[key][i]
#     argSentence = create_command(searchArgs)
#     print(COMMAND + argSentence)
#     os.system(COMMAND + argSentence)
# messagebox.showinfo("完成", "程序已成功运行完成！")
